@@ -1,8 +1,24 @@
+from typing import Iterable, Optional
 from django.db import models
-from django.contrib.contenttypes.models import ContentType
+from django.contrib.contenttypes.models import ContentType, ContentTypeManager
 from django.utils.translation import gettext_lazy as _
 from .resolver import active_resolver, _ComputedFieldsModelBase
 
+__all__ = [
+    'ComputedFieldsModel',
+    'computed',
+    'precomputed',
+    'compute',
+    'update_computedfields',
+    'update_dependent',
+    'preupdate_dependent',
+    'has_computedfields',
+    'get_computedfields',
+    'is_computedfield',
+    'get_contributing_fks',
+    'ComputedFieldsAdminModel',
+    'ContributingModelsModel'
+]
 
 class ComputedFieldsModel(_ComputedFieldsModelBase, models.Model):
     """
@@ -14,8 +30,14 @@ class ComputedFieldsModel(_ComputedFieldsModelBase, models.Model):
     class Meta:
         abstract = True
 
-    def save(self, force_insert=False, force_update=False, using=None,
-             update_fields=None, skip_computedfields=False):
+    def save(
+        self,
+        force_insert: bool = False,
+        force_update: bool = False,
+        using: Optional[str] = None,
+        update_fields: Optional[Iterable[str]] = None,
+        skip_computedfields: bool = False
+    ) -> None:
         """
         Overloaded to update computed field values before writing to the database.
 
@@ -28,7 +50,12 @@ class ComputedFieldsModel(_ComputedFieldsModelBase, models.Model):
         """
         if not skip_computedfields:
             update_fields = update_computedfields(self, update_fields)
-        return super(ComputedFieldsModel, self).save(force_insert=force_insert, force_update=force_update, using=using, update_fields=update_fields)
+        return super(ComputedFieldsModel, self).save(
+            force_insert=force_insert,
+            force_update=force_update,
+            using=using,
+            update_fields=update_fields
+        )
 
 
 # some convenient access mappings
@@ -46,12 +73,8 @@ compute = active_resolver.compute
 update_computedfields = active_resolver.update_computedfields
 #: Convenient access to :meth:`update_dependent<.resolver.Resolver.update_dependent>`.
 update_dependent = active_resolver.update_dependent
-#: Convenient access to :meth:`update_dependent_multi<.resolver.Resolver.update_dependent_multi>`.
-update_dependent_multi = active_resolver.update_dependent_multi
 #: Convenient access to :meth:`preupdate_dependent<.resolver.Resolver.preupdate_dependent>`.
 preupdate_dependent = active_resolver.preupdate_dependent
-#: Convenient access to :meth:`preupdate_dependent_multi<.resolver.Resolver.preupdate_dependent_multi>`.
-preupdate_dependent_multi = active_resolver.preupdate_dependent_multi
 
 # helper
 #: Convenient access to :meth:`has_computedfields<.resolver.Resolver.has_computedfields>`.
@@ -64,7 +87,7 @@ is_computedfield = active_resolver.is_computedfield
 get_contributing_fks = active_resolver.get_contributing_fks
 
 
-class ComputedModelManager(models.Manager):
+class ComputedModelManager(ContentTypeManager):
     def get_queryset(self):
         objs = ContentType.objects.get_for_models(
             *active_resolver.computed_models.keys()).values()
@@ -88,7 +111,7 @@ class ComputedFieldsAdminModel(ContentType):
         ordering = ('app_label', 'model')
 
 
-class ModelsWithContributingFkFieldsManager(models.Manager):
+class ModelsWithContributingFkFieldsManager(ContentTypeManager):
     def get_queryset(self):
         objs = ContentType.objects.get_for_models(
             *active_resolver.get_contributing_fks().keys()).values()
